#!/usr/bin/python

import re


hotfix_pattern = re.compile(r'^\d\d\.\d\d_hotfix(\d\d)')

class PEMVersion:
	def __init__(self, version, hotfix=None):
		self.version = [int(x) for x in version.split('.')]


		if type(hotfix) in (int,):
			self.hotfix = hotfix
		elif hotfix is None:
			self.hotfix = None
		else:
			hfmatch = hotfix_pattern.match(hotfix)
			if hfmatch is None:
				self.hotfix = 0
			else:
				self.hotfix = int(hfmatch.group(1))

	def __cmp__(self, rhs):
		if type(rhs) in (str, unicode):
			to_cmp = PEMVersion(rhs)
		else:
			to_cmp = rhs

		if self.version == to_cmp.version:
			return cmp(self.hotfix, to_cmp.hotfix)
		else:
			return cmp(self.version, to_cmp.version)
	

	def __str__(self):
		rv = 'POA %s' % ('.'.join([str(x) for x in self.version]))
		if self.hotfix is not None:
			rv += ' hotfix %02d' % self.hotfix
		return rv
	
	def __repr__(self):
		return "<%s>" % self


	def version_same(self, rhs):
		if type(rhs) in (str, unicode):
			to_cmp = PEMVersion(rhs)
		else:
			to_cmp = rhs
		return to_cmp.version == self.version

def getPEMVersion():
	import uSysDB
	con = uSysDB.connect()

	cur = con.cursor()
	cur.execute("SELECT build, version FROM version_history WHERE install_date = (SELECT MAX(install_date) FROM version_history)");
	row = cur.fetchone()
	if not row:
		raise Exception("There is no any version registered")
	
	version, build = row[1], row[0]

	cur.execute("SELECT MAX(name) FROM hotfixes WHERE build = %s", build)

	return PEMVersion(version, cur.fetchone()[0])

def getSCVersion(name):
	import uSysDB
	con = uSysDB.connect()
	cur = con.cursor()

	if uSysDB.hamn_schema():
		cur.execute("SELECT version FROM service_classes WHERE name = %s", name)
	else:
		cur.execute(
			"SELECT version FROM packages p JOIN components c ON (p.pkg_id = c.pkg_id) "
			"JOIN service_classes sc ON (sc.component_id = c.component_id) WHERE sc.name = %s",
			name)

	row = cur.fetchone()
	if not row:
		raise Exception("Service controller '%s' is not installed" % name)

	return PEMVersion(row[0])

patterns = [
(re.compile(r"^(\d{2}.\d{2})(_hotfix\d{2})?(_KB\d{4,})?$"), "_hotfix", "_KB"),
(re.compile(r"^(\d{2}.\d{2})(_update\d{2})(_KB\d{4,})?$"), "_update", "_KB"),
(re.compile(r"^(poa-\d\.\d-\d+)(_update\d{2})?(_KB\d{4,})?$"), "_update", "_KB") ]

def parseVersion(version_str):
	version = update = hotfix = None
	for p, upd_prefix, hotfix_prefix in patterns:
		m = p.match(version_str)
		if m:
			version = m.group(1)
			if m.group(2):
				update = m.group(2)[len(upd_prefix):]
			if m.group(3):
				hotfix = m.group(3)[len(hotfix_prefix):]

	return version, update, hotfix

# convert build name (like "poa-5.2-23", "poa-5.2-435_update01") into
# descriptive product name and version like ("POA", "5.2"), ("POA", "5.0 update01")
def parseProductVersion(version_str):
	product = version = None
	pattern = re.compile(r"^(poa)-(\d\.\d+)-\w+_?(update\d{2})?")
	m = pattern.match(version_str)
	if m:
		product = m.group(1).upper()
		version = m.group(2)
		if m.group(3):
			version += " " + m.group(3)

	return product, version


poa_ver_pattern = re.compile(r'poa-(\d)\.(\d)-\d+')
oldver_pattern = re.compile(r'(\d{2})\.(\d{2})')
def cmpJustVer(v1, v2):
	for pattern in (poa_ver_pattern, oldver_pattern):
		v1m = pattern.match(v1)
		v2m = pattern.match(v2)
		if v1m:
			if not v2m:
				return 1
			maj1 = v1m.group(1)
			maj2 = v2m.group(1)
			if maj1 < maj2:
				return -1
			if maj1 > maj2:
				return 1
			min1 = v1m.group(2)
			min2 = v2m.group(2)

			if min1 < min2:
				return -1
			if min1 > min2:
				return 1
			return 0
		elif v2m:
			return -1

	return 0
	


def compareVersions(v1, v2):
	v1v, v1u, v1h = parseVersion(v1)
	v2v, v2u, v2h = parseVersion(v2)

	if not v1v or not v2v:
		return 0
	rv = cmpJustVer(v1v, v2v)
	if rv:
		return rv

	if not v1u:
		if not v2u:
			return 0
		else:
			return -1
	elif not v2u:
		return 1
	
	return cmp(v1u, v2u)

if __name__ == '__main__':
	import uConfig, sys
	config = uConfig.Config(uConfig.commonConfigArgs, sys.argv)
	if config.args:
		for arg in config.args:
			print parseVersion(arg)
	else:
		import uSysDB
		uSysDB.init(config)
		print getPEMVersion()
