# $Id: openapi.py 643120 2011-03-16 02:58:26Z akalinin $

import xmlrpclib as xmlrpc
from base64 import encodestring
import socket, errno, sys, time
from uLogging import notice, err, warn
import uDialog, uAction

class OpenAPISettings:
	host = 'localhost'
	port = 8440
	user = None
	password = None
	https = False

	def __str__(self):
		if self.user is None:
			return self.url()
		else:
			return "%s@%s" % (self.user, self.url())
	
	def url(self):
		if self.https:
			proto = 'https'
		else:
			proto = 'http'
		return '%s://%s:%s' % (proto, self.host, self.port)
	
class ConnectionRefused(Exception):
	def __init__(self):
		Exception.__init__(self, "Connection attempt to Open API refused. Please check if system is running")

_settings = OpenAPISettings()

def init(host = _settings.host, port = _settings.port, user=_settings.user, proto = _settings.https and 'https' or 'http', password = _settings.password):
	global _settings
	_settings.host = host
	_settings.port = port
	_settings.user = user
	if proto == 'http':
		_settings.https = False
	elif proto == 'https':
		_settings.https = True
	else:
		raise Exception("%s: invalid or not supported protocol" % proto)
	_settings.password = password


def initFromEnv():
	import uSysDB, uPEM, os
	con = uSysDB.connect()
	port = uPEM.getSCProperty(con, "OpenAPI", "xmlrpc.server.port")

	proto = 'http'
	user = passwd = None

	if uSysDB.table_exist(con, "openapi_config"):
		cur = con.cursor()
		cur.execute("SELECT require_auth, use_ssl FROM openapi_config")
		row = cur.fetchone()
		
		if row:
			if row[1] == 'y':
				proto = 'https'

			if row[0] == 'y':
				dummy, rootpath = uPEM.getMNInfo()
				propfile = open(os.path.join(rootpath, 'etc', 'Kernel.conf'))
				kernel_props = uPEM.readPropertiesFile(propfile)
				propfile.close()
				user = kernel_props['login']
				passwd = kernel_props['passwd']
	
	init(port = port, user=user, proto = proto, password = passwd)



class Transport(xmlrpc.SafeTransport):
	def __init__(self, config):
		self.config = config
		# do not have any idea what does that mean:
		self._use_datetime = 0

	def make_connection(self, host):
		if self.config.https:
			return xmlrpc.SafeTransport.make_connection(self, host)
		else:
			return xmlrpc.Transport.make_connection(self, host)


	def send_host(self, connection, host):
		if self.config.https:
			return xmlrpc.SafeTransport.send_host(self, connection, host)
		else:
			return xmlrpc.Transport.send_host(self, connection, host)

	def send_content(self, connection, request_body):
		connection.putheader("Content-Type", "text/xml")
		connection.putheader("Content-Length", str(len(request_body)))
		if self.config.user is not None:
			connection.putheader("Authorization: Basic %s" % encodestring("%s:%s" % (self.config.user, self.config.password)))
		connection.endheaders()
		if request_body:
			connection.send(request_body)



class OpenAPIError (Exception):
	def __init__(self, resp):
		Exception.__init__(self, "Error during OpenAPI call: %s" % resp['error_message'])
		self.error_message = resp['error_message']
		self.module_id = resp.get('module_id', 'unknown module')
		self.extype_id = resp.get('extype_id', 'unknown exception type')
		self.properties = resp.get('properties', dict())


def _transform_params(params):
	if type(params) == list:
		return [_transform_params(p) for p in params if p is not None]
	elif type(params) != dict:
		return params

	rv = {}
	for k, v in params.iteritems():
		if type(v) in (dict, list):
			rv[k] = _transform_params(v)
		elif v is not None:
			rv[k] = v
	return rv

class _Method:
	def __init__(self, api, name):
		self.name = name
		self.api = api
	def __getattr__(self, name):
		return _Method(self.api, '%s.%s' % (self.name, name))

	def __call__(self, **kwds):
		params = _transform_params(kwds)
		if self.api.txn_id is not None:
			params['txn_id'] = self.api.txn_id
		try:
			resp = getattr(self.api.server, self.name)(params)
		except socket.error, e:
			if type(e.args) == tuple and e.args[0] == errno.ECONNREFUSED:
				raise ConnectionRefused()
			else:
				raise

		if resp['status'] != 0:
			self.api.txn_id = None
			raise OpenAPIError(resp)
		return resp.get('result', None)

class OpenAPI:
	def __init__(self, settings = _settings, namespace = None):
		self.settings = settings
		self.server = xmlrpc.ServerProxy(self.settings.url(), transport = Transport(self.settings))
		self.txn_id = None
		self.namespace = namespace

		# shut up pychecker in most cases
		self.pem = _Method(self, 'pem')

	def checkedCall(self, method, params):
		resp = method(params)

		if resp['status'] != 0:
			raise OpenAPIError(resp)

		return resp

	def begin(self, txn_id = None, request_id = None):
		params = {}
		if txn_id is not None:
			params['txn_id'] = txn_id
		if request_id is not None:
			params['request_id'] = request_id
		try:
			resp = self.server.txn.Begin(params)
		except socket.error, e:
			if type(e.args) == tuple and e.args[0] == errno.ECONNREFUSED:
				raise Exception("Connection attempt to Open API refused. Please check if system is running")
			else:
				raise

		if resp['status'] != 0:
			raise OpenAPIError(resp)
		self.txn_id = resp['result']['txn_id']
	
	def beginRequest(self, request_id_cand = 'req'):
		tc = 0
		prefix = ''
		request_id = None
		while request_id is None:
			try :
				begin_request_id = prefix + request_id_cand
				# 'request_id' is used for composing task group name in TaskManager controller;
				# task group name is stored in data base in column with type 'character varying(256)'
				begin_request_id = begin_request_id[:256]
				self.begin(request_id = begin_request_id)
				request_id = begin_request_id
			except OpenAPIError, e:
				if e.module_id == 'OpenAPI' and e.extype_id == 3001:
					tc += 1
					prefix = '%d_' % tc
				else:
					raise
		return request_id
			


	def commit(self):
		if self.txn_id is None:
			raise Exception("No transaction in progress")
		try:
			self.server.txn.Commit({'txn_id': self.txn_id})
		except socket.error, e:
			if type(e.args) == tuple and e.args[0] == errno.ECONNREFUSED:
				raise Exception("Connection attempt to Open API refused. Please check if system is running")
			else:
				raise
		self.txn_id = None

	def rollback(self):
		if self.txn_id is None:
			raise Exception("No transaction in progress")
		self.server.txn.Rollback({'txn_id': self.txn_id})
		self.txn_id = None

	def __getattr__(self, name):
		if self.namespace is None:
			nm = name
		else:
			nm = '%s.%s' % (self.namespace, name)

		return _Method(self, nm)
	
	def __repr__(self):
		if self.txn_id is None:
			txn = "no transaction"
		else:
			txn = "txn '%s'" % self.txn_id

		return "<Open API at %s, %s>" % (self.settings, txn)


def waitRequestComplete(request_id, operation, doWait):
	waiting = False
	api = OpenAPI()
	default_tried = False
	while True:
		request_status = api.pem.getRequestStatus(request_id = request_id)
		if request_status['request_status'] == 1:
			if not waiting:
				notice("%s (request %s) is not completed yet.", operation, request_id)
		elif request_status['request_status'] == 2:
			err("%s operation failed", operation)
			for e in request_status['status_messages']:
				err('%s', e)

		else:
			notice("%s finished successfully", operation)
			return

		if doWait and not waiting:
			notice("Waiting for %s completion (Interrupt to abort)", operation)
			waiting = True
		if waiting:
			try:
				sys.stderr.write('.')
				time.sleep(3)
			except KeyboardInterrupt, e:
				doWait = False
		if doWait:
			continue

		default_error_action = uAction.get_default_error_action()
		if default_error_action and not default_tried:
			action = default_error_action
			default_tried = True
		else:
			action = uDialog.askVariants('What should I do', '(A)bort', ['(A)bort', '(R)echeck', '(I)gnore', '(W)ait'])
		if action in ('(A)bort', 'abort'):
			raise Exception('%s failed', operation)
		elif action in ('(I)gnore', 'ignore'):
			return
		elif action in ('(W)ait'):
			doWait = True
		elif action in ('(R)echeck', 'retry'):
			waiting = doWait = False
