__rcs_id__ = """$Id: uDBValidator.py 643120 2011-03-16 02:58:26Z akalinin $"""

__pychecker__ = "unusednames=__rcs_id__,dummy"

import uDBSchema, uDBTypes, uSysDB, uManifestParser, uPDLDBSchema, cStringIO

from uDBErrors import *
from uAction import progress


def validateDatabase():
	progress.do("gathering schema information from manifests")
	con = uSysDB.connect()
	cur = con.cursor()
	cur.execute("""SELECT sc.name, data FROM package_body p
	JOIN components c ON (c.pkg_id = p.pkg_id)
	JOIN sc_instances si ON (si.component_id = c.component_id)
	JOIN service_classes sc ON (sc.sc_id = si.sc_id)
	WHERE sc.name NOT IN ('pleskd', 'vzpemagent')""")

	schemas=dict()
	manifest_tables = dict()
	for row in cur.fetchall():
		scname = row[0]
		schemas[scname] = dict()
		data = row[1]
		data = str(data)
		xml_text = uManifestParser.unsign_package(cStringIO.StringIO(data))
		for table in uPDLDBSchema.getSchemaFromString(xml_text):
			schemas[scname][table.name] = table
			manifest_tables[table.name] = table

	progress.done()

	progress.do("gathering real database schema")
	tablist = uDBSchema.getTableList(con)
	tables = dict()
	for tname in tablist:
		tables[tname] = uDBSchema.getTable(tname, con)

	progress.done()

	validator = DBValidator()
	progress.do("checking for missing tables")
	missing = validator.searchMissingTables(manifest_tables, tables)
	progress.done()
	progress.do("validating table structure")
	invalid = validator.validateTables(manifest_tables, tables)
	progress.done()
	return missing, invalid

class DBValidator:
	# search redundant tables
	# @schemas - (schema name -> Table object) mapping
	# @tables - (table name -> Table object) mapping
	def searchRedundantTables(self, schemas, tables):
		errors = []

		for table_name in tables.keys():
			if not schemas.has_key(table_name):
				errors.append(DBErrorRedundantTable(tables[table_name]))
		return errors

	# search missing tables
	# @schemas - (schema name -> Table object) mapping 
	# @tables - (table name -> Table object) mapping
	def searchMissingTables(self, schemas, tables):
		errors = []

		for schema_name in schemas.keys():
			if not tables.has_key(schema_name):
				errors.append(DBErrorMissingTable(schemas[schema_name]))
		return errors

	# validate tables
	# @schemas - (schema name -> Table object) mapping
	# @tables - (table name -> Table object) mapping
	def validateTables(self, schemas, tables):
		errors = []

		for schema_name in schemas.keys():
			if tables.has_key(schema_name):
				errors += self._validateColumns(schemas[schema_name], tables[schema_name])
				errors += self._validateIndexes(schemas[schema_name], tables[schema_name])
				errors += self._validateConstraints(schemas[schema_name], tables[schema_name])
		return errors

	# validate table columns
	# @schema - Table object
	# @table - Table object
	def _validateColumns(self, schema, table):
		errors = []

		table_columns = {}
		for column in table.columns:
			table_columns[column.name] = column

		schema_columns = {}
		for schema_column in schema.columns:
			schema_columns[schema_column.name] = schema_column
			if not table_columns.has_key(schema_column.name):
				errors.append(DBErrorMissingColumn(table, schema_column))
			else:
				table_column = table_columns[schema_column.name]

				if not isinstance(table_column.type, type(schema_column.type)):
					errors.append(DBErrorNonconformingColumnType(table, table_column, schema_column))
				if table_column.nullable and not schema_column.nullable:
					errors.append(DBErrorColumnMissingNotNull(table, table_column))
				if not table_column.nullable and schema_column.nullable:
					errors.append(DBErrorColumnRedundantNotNull(table, table_column))

				if table_column.default is not None or schema_column.default is not None:
					if table_column.default is None:
						errors.append(DBErrorColumnMissingDefault(table, table_column, schema_column.default))
					elif schema_column.default is None:
						errors.append(DBErrorColumnRedundantDefault(table, table_column))
					else:
						if not isinstance(table_column.default, uDBTypes.DefSysDate):
							if not uDBSchema.exprsEq(table_column.default, schema_column.default):
								errors.append(DBErrorColumnNonconformingDefault(table, table_column, schema_column.default))
						elif not isinstance(schema_column.default, uDBTypes.DefSysDate):
							errors.append(DBErrorColumnNonconformingDefault(table, table_column, schema_column.default))

		for table_column in table.columns:
			if not schema_columns.has_key(table_column.name):
				errors.append(DBErrorRedundantColumn(table, table_column))
		return errors

	# validate table indexes
	# @schema - Table object
	# @table - Table object
	def _validateIndexes(self, schema, table):
		errors = []

		schema_indexes = {}
		for schema_index in schema.indexes:
			schema_indexes[schema_index.name] = schema_index

		table_indexes = {}
		for table_index in table.indexes:
			table_indexes[table_index.name] = table_index
			if not schema_indexes.has_key(table_index.name):
				errors.append(DBErrorRedundantIndex(table, table_index))

		for schema_index in schema.indexes:
			if not table_indexes.has_key(schema_index.name):
				errors.append(DBErrorMissingIndex(table, schema_index))
			else:
				table_index = table_indexes[schema_index.name]
				if len(table_index.columns) != len(schema_index.columns) or table_index.columns.sort() != schema_index.columns.sort():
					errors.append(DBErrorNonconformingIndex(table, table_index, schema_index))
		return errors

	# validate table constraints
	# @schema - Table object
	# @table - Table object
	def _validateConstraints(self, schema, table):
		errors = []

		schema_pk = None
		table_pk = None
		
		schema_constraints = {}
		for constraint in schema.constraints:
			schema_constraints[constraint.name] = constraint

		table_constraints = {}
		for constraint in table.constraints:
			table_constraints[constraint.name] = constraint
			if isinstance(constraint, uDBSchema.PrimaryKey):
				table_pk = constraint
			else:
				if not schema_constraints.has_key(constraint.name):
					if isinstance(constraint, uDBSchema.UniqueKey):
						errors.append(DBErrorRedundantUnique(table, constraint))
					elif isinstance(constraint, uDBSchema.ForeignKey):
						errors.append(DBErrorRedundantForeignKey(table, constraint))
					elif isinstance(constraint, uDBSchema.Check):
						errors.append(DBErrorRedundantCheck(table, constraint))

		for constraint in schema.constraints:
			if isinstance(constraint, uDBSchema.PrimaryKey):
				schema_pk = constraint
			else:
				if not table_constraints.has_key(constraint.name):
					if isinstance(constraint, uDBSchema.UniqueKey):
						errors.append(DBErrorMissingUnique(table, constraint))
					elif isinstance(constraint, uDBSchema.ForeignKey):
						errors.append(DBErrorMissingForeignKey(table, constraint))
					elif isinstance(constraint, uDBSchema.Check):
						errors.append(DBErrorMissingCheck(table, constraint))
				else:
					table_constraint = table_constraints[constraint.name]

					if isinstance(constraint, uDBSchema.UniqueKey):
						if len(table_constraint.columns) != len(constraint.columns) or table_constraint.columns.sort() != constraint.columns.sort():
							errors.append(DBErrorNonconformingUnique(table, table_constraint, constraint))
					elif isinstance(constraint, uDBSchema.ForeignKey):
						report = False

						if table_constraint.reftable != constraint.reftable:
							report = True

						if len(table_constraint.columns[0]) != len(constraint.columns[0]):
							report = True
						else:
							table_fk = map(lambda x, y: (x, y), table_constraint.columns[0], table_constraint.columns[1]).sort(lambda x, y: cmp(x[0], y[0]))
							schema_fk = map(lambda x, y: (x, y), constraint.columns[0], constraint.columns[1]).sort(lambda x, y: cmp(x[0], y[0]))
							if table_fk != schema_fk:
								report = True

						if table_constraint.ondelete != constraint.ondelete or table_constraint.onupdate != constraint.onupdate:
							report = True

						if report:
							errors.append(DBErrorNonconformingForeignKey(table, table_constraint, constraint))
					elif isinstance(constraint, uDBSchema.Check):
						if not uDBSchema.exprsEq(table_constraint.expression, constraint.expression):
							errors.append(DBErrorNonconformingCheck(table, table_constraint, constraint))

		if table_pk is not None or schema_pk is not None:
			if table_pk is None:
				errors.append(DBErrorMissingPrimaryKey(table, schema_pk))
			elif schema_pk is None:
				errors.append(DBErrorRedundantPrimaryKey(table, table_pk))
			elif table_pk.name != schema_pk.name:
				errors.append(DBErrorNonconformingPrimaryKey(table, table_pk, schema_pk))
			elif len(table_pk.columns) != len(schema_pk.columns) or table_pk.columns.sort() != schema_pk.columns.sort():
				errors.append(DBErrorNonconformingPrimaryKey(table, table_pk, schema_pk))
		return errors

