from uBuild import *
from uManifestParser import *
from uAction import *
from uLogging import *
from uRPMUtils import RPMInfo
import MDLReader

import os
import sys
import os.path
import re
import tempfile
import shutil
import tarfile
import md5

tarball_pattern = re.compile(r'.+\.(tar\.gz|tgz)$');
manifest_pattern = re.compile(r'.+\.pdl\.asc$')

rpm_pattern = re.compile(r'.+\.rpm$')
template_pattern = re.compile(r'.+-(template|ez)-.+')
exe_pattern = re.compile(r'.+\.(exe|msi|efd|dll|pdf)$')
mdl_pattern = re.compile(r'.+\.mdl$')
platform_dir_pattern = re.compile(r'^.*(Fedora|CentOS|RHES|RHEL)[-\\/]([\d\.]+)')

def update_filemap(fmap, dirname, fnames):
	for fn in fnames:
		fp = os.path.join(dirname, fn)
		if os.path.isfile(fp):
			fmap[fp[1:]] =  md5.new(open(fp, 'rb').read()).hexdigest()

def _collect_rpm_content(path):
	log("-- calculating RPM contents checksums")
	tempdir = tempfile.mkdtemp()
	realrpmpath = os.path.realpath(path)
	curdir = os.getcwd()
	filemap = {}
	try :
		os.chdir(tempdir)
		## W/A for RHEL5 bug "rpm2cpio fails on rhel-6 rpms" https://bugzilla.redhat.com/show_bug.cgi?id=602423
		## On deliver.vt.sw.ru
		## 1) yum install xz
		## 2) create /usr/bin/rpm2cpio.sh
		if os.system('prog="rpm2cpio"; if [ -f "/usr/bin/rpm2cpio.sh" ]; then prog="rpm2cpio.sh"; fi; $prog %s | cpio -d -i' % realrpmpath):
			raise Exception, 'Failed to unpack %s' % realrpmpath
		os.path.walk('.', update_filemap, filemap)
	finally:
		os.chdir(curdir)
		shutil.rmtree(tempdir)

	checksum = md5.new(file(path).read()).hexdigest()
	return BuiltPackageContent(os.path.basename(path), checksum, filemap)

def parseRPM(path, platform, calculate_checksums = True):
	log("Processing %s", path)
	rpminfo = RPMInfo(path)
	arch = rpminfo.arch

	if arch != 'x86_64':
		arch = 'i386' # no stupid i586 (XXX: what about noarch?)

	if platform:
		pem_platform = Platform(platform[0], platform[1], arch)
	else:
		pem_platform = Platform('any', 'any', 'any')

	rpm = RPM(rpminfo.name, pem_platform, os.path.basename(path))

	if calculate_checksums and sys.platform != 'win32':
		content = _collect_rpm_content(path)
	else:
		content = None

	return BuiltRPM(rpm, rpminfo.version, rpminfo.release, content)

def add_files(build, dirname, files):
	pmatch = platform_dir_pattern.match(dirname)
	if pmatch:
		guessed_platform = (pmatch.group(1), pmatch.group(2))
	else:
		guessed_platform = None

	if os.path.basename(dirname) == 'bin' and guessed_platform:
		if not build.platform_binaries.has_key(guessed_platform):
			build.platform_binaries[guessed_platform] = []
		build.platform_binaries[guessed_platform] += [os.path.join(dirname, x) for x in files]

	for fn in files:
		fullname = os.path.join(dirname, fn)
		if os.path.islink(fullname) and os.path.isdir(fullname):
			notice("Walking down symlinked directory %s", fullname)
			os.path.walk(fullname, add_files, build)
		elif manifest_pattern.match(fn):
			try:
				pkg_info = get_package(fullname)
			except Exception, e:
				err("%s while processing %s", e, fullname)
				raise
			pkg_info.manifest_file = fullname
			pkg_info.topdir = build.topdir
			build.add_package(pkg_info, build.fail_on_duplicates)
		elif tarball_pattern.match(fn):
			if not build.tarballs_locations.has_key(fn):
				build.tarballs_locations[fn] = []
			build.tarballs_locations[fn].append(dirname)
		elif fn == 'install_win_sn.exe':
			build.windows_update.installer_path = fullname
			build.add_file("win32", fn, dirname)
		elif fn.lower() == 'asyncexec.exe':
			build.windows_update.async_exec_path = fullname
			build.add_file("win32", fn, dirname)
		elif rpm_pattern.match(fn):
			if template_pattern.match(fn):
				log("Processing template %s", fullname)
				build.add_template(fullname)
			else:
				if not build.rpm_files.has_key(fn):
					build.rpm_files[fn] = {}
				rpm = parseRPM(fullname, guessed_platform, build.calculate_checksums)
				build.rpm_files[fn][guessed_platform] = dirname
				build.add_rpm(rpm)

			if not guessed_platform:
				notice("Assuming any-any-any platform for %s", fullname)

			build.add_file(guessed_platform, fn, dirname)
		elif fn == 'update.udl2':
			build.instructions.append(fullname)
		elif exe_pattern.match(fn) or fn == 'pem_mn.msi':
			build.add_file("win32", fn, dirname)
		elif mdl_pattern.match(fn):
			build.add_module(MDLReader.readModule(fullname))

def rmdir(dir):
	try:
		shutil.rmtree(dir)
	except:
		pass

def getBuildFromFS(path, buildname, conf=None, fail_on_duplicates = False, calculateChecksums = True):
	rv = Build(buildname, conf)
	rv.tarballs_locations = {}
	rv.instructions = []
	rv.rpm_files = {}
	rv.fail_on_duplicates = fail_on_duplicates
	rv.calculate_checksums = calculateChecksums

	if conf and hasattr(conf, "cache") and conf.cache:
		tempfile.tempdir = conf.cache
	else:
		tempfile.tempdir = tempfile.mkdtemp()
		import atexit
		atexit.register(rmdir, tempfile.tempdir)

	curdir = os.getcwd()
	rv.topdir = path
	try:
		progress.do("processing %s", path)
		os.chdir(path)
		os.path.walk(".", add_files, rv)
		progress.done()
	finally:
		os.chdir(curdir)
	return rv

__all__ = [ "getBuildFromFS" ]
