import uUtil, uBuild, os, time, sys, re, socket, uLogging, errno
from uAction import *

def initYum(rpms_directory, cache_directory):
	# here we
	#  - add RPMS directory from POA distribution to YUM, as REPO
	#  - and check that YUM works
	import yum

	try:
		from yum.rpmtrans import RPMBaseCallback

		# here we define class inside function, to properly handle yum absence and display user friendly error message
		class PAYumCliCallBack(RPMBaseCallback):
		    def __init__(self):
			RPMBaseCallback.__init__(self)
			self.lastpkg = None

		    def event(self, package, action, te_current, te_total, ts_current, ts_total):
			if type(self.lastpkg) != type(package) or self.lastpkg != package:
				uLogging.notice('%s: %s [%s/%s]' % (self.action[action], package, ts_current, ts_total))
				self.lastpkg = package

		    def scriptout(self, package, msgs):
			pass

		class PAYumBase(yum.YumBase):
			def __init__(self):
				yum.YumBase.__init__(self)

				self.initYumBase()
				self.predefinedCallback = PAYumCliCallBack()

			def initYumBase(self):
				# TODO is there any other way to init logging in YUM to avoid this jumping around?
				# disable logging to avoid output to console (this can break snackUI)
				self.doConfigSetup(debuglevel=-1,errorlevel=-1)

				# conf and logging is initialized after first read
				# so we need init conf and only after that replace logging handlers
				self.conf.assumeyes = True
				self.conf.rpm_check_debug = False

				# replace YUM logging to console with logging to installer log
				uHandler = uLogging.Handler()

				import logging
				logger = logging.getLogger("yum")
				for handler in logger.handlers:
					logger.removeHandler(handler)
				logger.addHandler(uHandler)
				logger.setLevel(logging.DEBUG)
				logger = logging.getLogger("yum.verbose")
				for handler in logger.handlers:
					logger.removeHandler(handler)
				logger.addHandler(uHandler)
				logger.setLevel(logging.DEBUG)

			def buildAndProcessTransaction(self):
				res, msg = self.buildTransaction()
				if res not in [0, 2]: # 0 - empty transaction, 2 - OK
					raise Exception("Failed to resolve dependencies. Check that OS YUM repository is enabled or install required packages manually:\n %s" % ("".join(msg)))
				self.processTransaction(rpmDisplay=self.predefinedCallback)

			def cleanAll(self):
				yb.cleanPackages()
				yb.cleanHeaders()
				yb.cleanMetadata()
				yb.cleanSqlite()
			
		yb = PAYumBase()

		# why YUM doesn't handle already installed package by itself?
		if not yb.isPackageInstalled('createrepo'):
			uLogging.notice("yum install createrepo")
			yb.install(name='createrepo')
			yb.buildAndProcessTransaction()

		uLogging.notice("Creating YUM repository in %s" % rpms_directory)
		uUtil.execCommand(["createrepo", rpms_directory])

		# cannot reuse already created YumBase, it seems it cannot 'see' new repos after processTransaction
		yb = PAYumBase()

		newrepo = yum.yumRepo.YumRepository("poa-local-distrib")
		newrepo.name = "poa-local-distrib"
		newrepo.baseurl = "file://" + rpms_directory
		newrepo.basecachedir = cache_directory

		# TODO cannot be enabled until we have notsigned RPMs in pem-contrib
		newrepo.gpgcheck = 0

		yb.repos.add(newrepo)
		yb.repos.enableRepo(newrepo.id)

		# to avoid stale caching problems it's always a good idea to clean up YUM caches during update & installation
		yb.cleanAll()
	except ImportError, e:
		# YUM 2 doesn't have rpmtrans and is not supported for new installations
		# but for backwards compatibility we must support some of its functionality
		uLogging.notice("Import error occured during YUM intialization: %s" % e)
		uLogging.notice("Using compatibility mode, which is supported only for upgrade. Installation on RHEL/CentOS older than 5 is not possible.")

		yb = yum.YumBase()

		# to avoid stale caching problems it's always a good idea to clean up YUM caches during update & installation
		yb.cleanPackages()
		yb.cleanHeaders()
		yb.cleanMetadata()
		yb.cleanSqlite()

	repos_info = yb.doRepoSetup()
	if repos_info is not None:
		uLogging.debug("Enabled YUM repositories:")
		enabledRepos = repos_info.listEnabled()
		for r in enabledRepos:
			uLogging.debug("%s: %s (%s)" % (r.name, r.baseurl, r.basecachedir))
			r.ready()

		if len([r for r in enabledRepos if r.name != 'poa-local-distrib']) == 0:
			uLogging.notice("There are no enabled YUM repositories on this host. Consider checking OS YUM repository configuration.")

	return yb

def translatePlatformForRPMManagement(platform):
	opsys, osver, oparch = platform.os, platform.osver, platform.arch

        if opsys == 'CentOS':
                if osver in ('3', '4'):
                        opsys = 'RHES'
                else:
                        opsys = 'RHEL'

	return (opsys, osver, oparch)

def isPostgreSQLStarted():
	out, err, status = uUtil.readCmdExt(["/sbin/service", "postgresql-9.0", "status"])
	return status == 0
	
def stopPostgreSQL():
	uUtil.execCommand("service postgresql-9.0 stop")

def startPostgreSQL():
	uUtil.execCommand("service postgresql-9.0 start")

	count = 300
	status = 1
	hostname = socket.gethostname()
	while status and count:
		dummy, dummy, status = uUtil.readCmdExt(["/usr/pgsql-9.0/bin/psql", "-Upostgres", "-dtemplate1",  "-h", hostname, "-tc", "SELECT 1"])
		time.sleep(1)
		count -= 1


def inVZContainer():
	return os.path.exists('/proc/user_beancounters') and not os.path.exists('/etc/virtuozzo-release')

redhat_release_pattern = re.compile(r"(.+) release (\d+).*")

def determineRedHatFlavour(name, filename):
	f = open(filename)
	release_str = f.read()
	f.close()
	rmatch = redhat_release_pattern.match(release_str)
	osname = rmatch.group(1)
	osver = rmatch.group(2)
	if osname == 'CentOS':
		name = 'CentOS'
	elif name == 'RHES' and osver >= '5':
		name = 'RHEL'
        return name, osver

known_flavour_files = [
	( "/etc/fedora-release", 'Fedora', determineRedHatFlavour ),
	( "/etc/altlinux-release", 'ALTLinux', determineRedHatFlavour ),
	( "/etc/redhat-release", 'RHES', determineRedHatFlavour )
];

def determinePlatform():
	arch = uUtil.readCmd(["uname",  "-i"]).strip()
	for filename, opsys, fun  in known_flavour_files:
		if os.path.exists(filename):
			opsys, ver = fun(opsys, filename)
			return uBuild.Platform(opsys, ver, arch)
	raise Exception("Failed to determine OS")

mempattern = re.compile("MemTotal:\s+(\d+) kB")
def getMemTotal():
	f = open("/proc/meminfo")
	try:
		all_mem = sh_mem = None
		for line in f:
			lm = mempattern.match(line)
			if lm:
				all_mem = int(lm.group(1)) * 1024

		if inVZContainer():
			f.close()
			privvmpages = shmpages = None
			f = open("/proc/user_beancounters")
			for line in f.readlines():
				cols = line.split()
				if cols and len(cols) > 4:
					if cols[0] == 'privvmpages':
						privvmpages = int(cols[4])
					elif cols[0] == 'shmpages':
						shmpages = int(cols[4])



			if shmpages:
				sh_mem = min(all_mem, shmpages * 4096)
			if privvmpages:
				all_mem = min(all_mem, privvmpages * 4096)



		if not sh_mem:
			sh_mem = all_mem
		if all_mem:
			if determinePlatform().arch != 'x86_64': # i386, i586 etc
				all_mem = min(all_mem, 4294967295)
				sh_mem = min(sh_mem, 4294967295)
			return all_mem, sh_mem

	finally:
		f.close()

	raise Exception("Cannot find MemTotal line in /proc/meminfo")



def adjustMemSettings(product_name, mem):
	def edit_sysctl_conf(infile, outfile, invz, mem):
		changed_all = changed_max = False
		for ln in infile:
			l = ln.split('=', 1)
			if len(l) == 2:
				p, v = l
				p = p.strip()
				if p == 'kernel.shmall':
					print >> outfile, 'kernel.shmall = %s' % mem
					changed_all = True
				elif p == 'kernel.shmmax':
					print >> outfile, 'kernel.shmmax = %s' % mem
					changed_max = True
				elif invz and p == "net.ipv4.tcp_syncookies":
					print >> outfile, '#', ln.strip()
				else:
					outfile.write(ln)
			else:
				outfile.write(ln)
		if not changed_all or not changed_max:
			print >> outfile, "# Required by", product_name
		if not changed_all:
			print >> outfile, 'kernel.shmall = %s' % mem
		if not changed_max:
			print >> outfile, 'kernel.shmmax = %s' % mem
	uUtil.editFileSafe('/etc/sysctl.conf', edit_sysctl_conf, '/etc/sysctl.conf.pasave', inVZContainer(), mem)
	uUtil.execCommand(["sysctl", "-p"])

ifconfig = "/sbin/ifconfig"

addr_pattern = re.compile(r".*inet addr:([.\d]+) .* Mask:([.\d]+)")

def listNetifaces():
	ifcfg_out = uUtil.readCmd([ifconfig])
	iface_name = ip = mask = None
	rv = []
	for line in ifcfg_out.splitlines():
		if line and not line[0].isspace(): # new interface
			if iface_name:
				if not ip:
					raise Exception("Failed to parse ifconfig output: No ip address for iface %s " % iface_name)
				if not mask:
					raise Exception("Failed to parse ifconfig output: No netmask for iface %s " % iface_name)
				rv.append((iface_name, ip, mask))


			iface_name = line.split()[0]
			ip = mask = None
		m = addr_pattern.match(line)
		if m:
			ip = m.group(1)
			mask = m.group(2)
	rv.append((iface_name, ip, mask))
	return rv

tarball_pattern = re.compile(r'.+\.(tar\.gz|tgz)$');
def copy_tarballs(to, dirname, fnames):
	for fn in fnames:
		if tarball_pattern.match(fn):
			uUtil.ln_cp(os.path.join(dirname, fn), to)
httpd_conf_dir = "/etc/httpd/conf.d"

def createPPMSite(props):
	rootpath = props['rootpath']
	conf = open(os.path.join(httpd_conf_dir, "pa_tarballs.conf"), "w")
	tdir = os.path.join(rootpath, "install", "tarballs")
	if not tdir.endswith('/'):
		tdir += '/'
	conf.write("""Alias /tarballs/ "%(dir)s"

<Directory "%(dir)s">
    Options Indexes MultiViews
    AllowOverride None
    Order allow,deny
    Allow from all
</Directory>
""" % { "dir" : tdir} )
	conf.close()
	try:
		os.makedirs(tdir)
	except OSError, e:
		if e.errno != errno.EEXIST:
			raise
	source_dir = props.get('source_dir', None)

	if source_dir is not None:
		os.path.walk(source_dir, copy_tarballs, tdir)

	uLogging.debug("%s", uUtil.readCmdExt(["service", "httpd", "stop"]))
	uUtil.execCommand(["service", "httpd", "start"])
