import uSysDB
from sets import Set

from uBuild import compare_versions
from uBuild import Platform

from uLogging import warn, notice, debug, err
from uAction import *
import uDialog, uPEM, uUtil, uNativeRepo, uHCL
import threading
import openapi
import os, re, sys
from uGenerics import first

try:
	import subprocess as sp
except ImportError:
	# for temporary backward compatibility with centos 4 and python 2.3
	import pasubprocess as sp

def init(newest_packages, plesk_root):
	global _newest_packages, _mn_plesk_root, _ppm_ctl_cmd
	_newest_packages, _mn_plesk_root = newest_packages, plesk_root

_newest_packages = None
_mn_plesk_root = None

_ppm_ctl_cmd = None

def ppm_ctl_cmd():
	global _mn_plesk_root, _ppm_ctl_cmd

	if _ppm_ctl_cmd is None:
		if _mn_plesk_root is None:
			dummy, _mn_plesk_root = uPEM.getMNInfo()
		_ppm_ctl_cmd = [os.path.join(_mn_plesk_root, 'bin', 'ppm_ctl'), '-b', '-q', '-f', os.path.join(_mn_plesk_root, 'etc', 'pleskd.props')]
	return _ppm_ctl_cmd

class PackageNotFound(Exception):
	def __init__(self, name, ctype, platform):
		Exception.__init__(self, "Cannot find %s-%s for %s" % (ctype, name, platform))

def getLatestPackages(name, ctype):
	if _newest_packages is None:
		# We're not initialized, it means that POST section is not in progress yet
		raise Exception, "Packages installation is allowed in POST phase only"
	rv = []
	for pf in _newest_packages:
		if _newest_packages[pf].get(ctype, {}).has_key(name):
			rv.append(_newest_packages[pf][ctype][name])
	return rv

def findSuitablePackage(host_id, name, ctype):
	con = uSysDB.connect()
	platform, rootpath = uPEM.getHostInfo(con, host_id)
	platforms = uPEM.getPlatformLine(con, platform)
	cur = con.cursor()
	if _mn_plesk_root is None:
		cur.execute("SELECT default_rootpath FROM hosts WHERE host_id = 1")
		init(None, cur.fetchone()[0])

	cur.execute(("SELECT p.pkg_id, p.version FROM packages p WHERE name = %%s AND ctype = %%s AND platform_id IN (%s)" % ','.join(['%s']*len(platforms))), [name, ctype] + [p.platform_id for p in platforms])
	pkgs = cur.fetchall()
	rv = None
	for p in pkgs:
		if rv is None or compare_versions(p[1], rv[1]) > 0:
			rv = p
	if rv is None:
		raise PackageNotFound(name, ctype, platform)
	return rv


def installPackageToHost(host_id, name, ctype, properties = None):
	pkg_id, version = findSuitablePackage(host_id, name, ctype)
	notice("Installing %s-%s-%s (%s) to host %s", ctype, name, version, pkg_id, host_id)
	if properties is None:
		proplist = []
	else:
		proplist = [ ("%s=%s" % (x, properties[x]) ) for x in properties ]
	uUtil.execCommand(ppm_ctl_cmd() + ["install", str(host_id), str(pkg_id), "INSTALL"] + proplist)

def installPackageToHostAPI(host_id, name = None, ctype = None, pkg_id = None, properties = None, print_name = None):
	if pkg_id is None:
		pkg_id, version = findSuitablePackage(host_id, name, ctype)
	else:
		version = None

	if print_name is None:
		if name is None:
			print_name = 'package'
		elif ctype is None:
			print_name = name
		elif version is None:
			print_name = "%s-%s" % (ctype, name)
		else:
			print_name = "%s-%s-%s" % (ctype, name, version)


	notice("Installing %s(%s) to host %s", print_name, pkg_id, host_id)

	api = openapi.OpenAPI()
	if properties is None:
		proplist = []
	else:
		proplist = [ { "name" : x, "value" : properties[x] } for x in properties ]
	return api.pem.packaging.installPackageSync(host_id = host_id, package_id = pkg_id, properties = proplist)["component_id"]


class PkgOnHost:
	def __init__(self, row):
		self.component_id, self.host_id, self.name, self.ctype, self.version, self.hostname = row[0], row[1], row[2], row[3], row[4], row[5]

def _getPkgsByQuery(q, *params):
	con = uSysDB.connect()
	cur = con.cursor()
	cur.execute("SELECT c.component_id, c.host_id, p.name, p.ctype, p.version, h.primary_name " + q, *params)
	
	return [ PkgOnHost(row) for row in cur.fetchall() ]

def getDependencies(component_id):
	return _getPkgsByQuery("FROM component_dependencies d JOIN components c ON (c.component_id = d.dep_id) JOIN packages p ON (c.pkg_id = p.pkg_id) JOIN hosts h ON (h.host_id = c.host_id) WHERE d.component_id = %s", component_id)
	
def listInstalledPackages(name, ctype):
	return _getPkgsByQuery("FROM components c JOIN packages p ON (c.pkg_id = p.pkg_id) JOIN hosts h ON (c.host_id = h.host_id) WHERE p.name = %s AND p.ctype= %s", name, ctype)

def removeComponent(host_id, component_id, hostname="host", pkg_name = "package", removeDepends = None):
	progress.do("removing %s(id = %s) from %s(id = %s)", pkg_name, component_id, hostname, host_id)

	deps = getDependencies(component_id)

	while deps:
		warn("%s-%s-%s (%s) installed on %s(%d) depends on %s (%s)", deps[0].ctype, deps[0].name, deps[0].version, deps[0].component_id, deps[0].hostname, deps[0].host_id, pkg_name, component_id )
		if ((removeDepends is None) and uDialog.askYesNo("Shall I remove it also?", False)) or removeDepends:
			removeComponent(deps[0].host_id, deps[0].component_id, deps[0].hostname, deps[0].name, removeDepends)
		else:
			raise Exception("Cannot deinstall package, there are dependencies")
		deps = getDependencies(component_id)


	uUtil.execCommand(ppm_ctl_cmd() + ['deinstall', str(host_id), str(component_id)])
	progress.done()

def removePkgFromRepo(pkg_id, nicename='package'):
	progress.do("removing %s (%s) from repository", nicename, pkg_id)
	uUtil.execCommand(ppm_ctl_cmd() + ['remove', 'pkg', str(pkg_id)])
	progress.done()


def removePackage(name, ctype, host_id = None, removeDepends = None):
	""" Removes package from host or from the whole system """
	con = uSysDB.connect()

	cur = con.cursor()

	nicename = "%s-%s" % (ctype, name)

	if _mn_plesk_root is None:
		cur.execute("SELECT default_rootpath FROM hosts WHERE host_id = 1")
		init(None, cur.fetchone()[0])

	if host_id is None:
		cur.execute("SELECT h.host_id, h.primary_name, c.component_id, h.htype FROM packages p JOIN components c ON (c.pkg_id = p.pkg_id) JOIN hosts h ON (h.host_id = c.host_id) WHERE p.name = %s AND p.ctype = %s", (name, ctype))

		to_deinstall = [ (row[0], row[1], row[2], row[3]) for row in cur.fetchall()]
		cur.execute("SELECT pkg_id FROM packages WHERE name = %s AND ctype = %s", (name, ctype))
		to_remove = [row[0] for row in cur.fetchall()]
	else:
		cur.execute("SELECT h.host_id, h.primary_name, c.component_id, h.htype FROM packages p JOIN components c ON (c.pkg_id = p.pkg_id) JOIN hosts h ON (h.host_id = c.host_id) WHERE p.name = %s AND p.ctype = %s AND h.host_id = %s", (name, ctype, host_id))

		to_deinstall = [ (row[0], row[1], row[2], row[3]) for row in cur.fetchall()]
		to_remove = []
	
	if not to_deinstall and host_id:
		notice("%s is not installed on host %s, doing nothing", (nicename, host_id))
		return
	elif not to_deinstall:
		notice("%s is not installed anywhere, doing nothing", nicename)
	else:
		for td in to_deinstall:
			host_id, hostname, component_id, host_type = td
			stop_ui = False
			# we should stop UI on Windows as we have locked jars
			if ctype == 'cp':
				if host_type == 'w':
					stop_ui = True
				if host_type == 'n':
					platform, rootpath = uPEM.getHostInfo(con, host_id)
					if platform.os == 'Win32':
						stop_ui = True
			if stop_ui:
				stopped_successfully = (uPEM.stopUI(host_id) is not None)

			removeComponent(host_id, component_id, hostname, nicename, removeDepends)

			# start UI again only after successful stop, otherwise
			# do think that UI servers was intentionally stopped before
			if stop_ui and stopped_successfully:
				uPEM.startUI(host_id)
	
	if not to_remove:
		if not host_id:
			notice("There is no %s in package repository, nothing to remove", nicename)
	else:
		for pkg_id in to_remove:
			removePkgFromRepo(pkg_id, nicename)

def registerPlatform(platform, parent, con = None):
	own_con = False
	if con is None:
		own_con = True
		con = uSysDB.connect()
	cur = con.cursor()
	cur.execute(
		"SELECT 1 FROM platforms "
		"WHERE arch = %s AND opsys = %s AND osrel = %s",
		(platform.arch, platform.os, platform.osver))
	if cur.fetchone():
		return

	cur.execute(
		"SELECT r, platform_id FROM platforms "
		"WHERE arch = %s AND opsys = %s AND osrel = %s",
		(parent.arch, parent.os, parent.osver))
	parent_platform = cur.fetchone()
	if not parent:
		raise Exception("Parent platform %s-%s-%s not found" % (parent.arch, parent.os, parent.osver))
	r = parent_platform[0]
	parent_id = parent_platform[1]

	cur.execute("UPDATE platforms SET l = l + 2 WHERE l > %s", r)
	cur.execute("UPDATE platforms SET r = r + 2 WHERE r >= %s", r)
	cur.execute("INSERT INTO platforms(arch, opsys, osrel, parent_id, l, r) "
		    "VALUES (%s, %s, %s, %s, %s, %s)",
		    platform.arch, platform.os, platform.osver, parent_id, r, r + 1)

	if own_con:
		con.commit()

__magic_value = 0xdeadbee	# should be less than max signed 32-bit int

def __rebuildSubPlatforms(con, parent_id):
	cur = con.cursor();

	cur.execute("SELECT platform_id FROM platforms "
		    "WHERE parent_id = %s AND platform_id <> parent_id ORDER BY platform_id", parent_id)
	for row in cur.fetchall():
		platform_id = row[0]

		# re-fetch parent r (changed on every subplatform re-addition)
		cur.execute("SELECT r FROM platforms WHERE platform_id = %s", parent_id)
		row = cur.fetchone()
		if not row:
			raise Exception("Parent platform #%s not found" % (parent_id))
		r = row[0]

		# re-add subplatform
		cur.execute("UPDATE platforms SET l = l + 2 WHERE l <> %s AND l > %s", __magic_value, r)
		cur.execute("UPDATE platforms SET r = r + 2 WHERE r <> %s AND r >= %s", __magic_value, r)
		cur.execute("UPDATE platforms SET l = %s, r = %s WHERE platform_id = %s",
			    r, r + 1, platform_id)
		__rebuildSubPlatforms(con, platform_id)

def rebuildPlatformTree(con):
	cur = con.cursor()

	# get any-any-any id
	cur.execute("SELECT platform_id FROM platforms "
		    "WHERE arch = 'any' AND opsys = 'any' AND osrel = 'any'")
	row = cur.fetchone()
	if not row:
		raise Exception("Platform any-any-any not found")
	platform_id = row[0]

	# set l, r for any-any-any and rebuild subplatforms
	cur.execute("UPDATE platforms SET l = 1, r = 2 WHERE platform_id = %s", platform_id)
	cur.execute("UPDATE platforms SET l = %s, r = %s WHERE platform_id <> %s",
		    __magic_value, __magic_value, platform_id)
	__rebuildSubPlatforms(con, platform_id)

	# check for dangling platforms (that are not children of any-any-any)
	cur.execute("SELECT platform_id, arch, opsys, osrel FROM platforms WHERE l = %s OR r = %s",
		    __magic_value, __magic_value)
	for row in cur.fetchall():
		warn("Dangling platform #%d (%s-%s-%s)", row[0], row[1], row[2], row[3])

def getNewestPackages(builds, filt=None):
	progress.do("searching for latest packages to install")
	rv = {}
	for build in builds:
		for pkg in build.contents.values():
			platform = pkg.package.platform
			if not rv.has_key(platform):
				rv[platform] = {}
			ctype = pkg.package.ctype
			if not rv[platform].has_key(pkg.package.ctype):
				rv[platform][ctype] = {}
			name = pkg.package.name
			if not (rv[platform][ctype].has_key(pkg.package.name)) or (compare_versions(rv[platform][ctype][name].version, pkg.version) <= 0):
				pkg.tarball_location = build.find_valid_tarball(pkg)
				if pkg.tarball_location:
					pkg.tarball_location = os.path.realpath(pkg.tarball_location)
				if filt is None or (pkg.package.ctype, pkg.package.name) in filt:
					rv[platform][ctype][name] = pkg
	progress.done()
	return rv

mass_import_out_pattern = re.compile("^([one]) (\\d+)\n?\r?$")

def __assignPkgIds(output, pkglist):
	i = 0
	line = output.readline()
	while line:
		if i <= len(pkglist):
			pkg = pkglist[i]
		outp = mass_import_out_pattern.match(line)
		if not outp:
			err("Cannot understand ppm_ctl output: '%s'", line)
		oldnew, pkg_id = outp.group(1), outp.group(2)
		pkg.pkg_id = int(pkg_id)
		if oldnew == 'n':
			pkg.old = False
		elif oldnew == 'o':
			pkg.old = True
		else:
			pkg.old = None
		i += 1
		line = output.readline()

def __showProgress(output, pkg_counter, full_output):
	line = output.readline()
	while line:
		pkg_counter.new_item(os.path.basename(line))
		full_output.append(line)
		line = output.readline()

def __packageImported(pkg):
	"""Return True if package with specific name, ctype, and platform exists in POA database. Version is ignored."""
	con = uSysDB.connect()
	cur = con.cursor()
	cur.execute("SELECT 1 FROM dual WHERE EXISTS (SELECT 1 FROM packages pk JOIN platforms pl ON (pk.platform_id = pl.platform_id) WHERE pk.name = %s AND pk.ctype = %s AND pl.arch = %s AND pl.opsys = %s AND pl.osrel = %s)", pkg.name, pkg.ctype, pkg.platform.arch, pkg.platform.os, pkg.platform.osver)
	return bool(cur.fetchall())

def __filterCustomPackages(pkglist):
	"""Filters out customer specific packages that were not imported before."""
	return [ p for p in pkglist if not p.package.is_custom or __packageImported(p.package) ]

def doMassImport(rootpath, pkglist, pkg_counter, topdir = None):
	command = [ os.path.join(rootpath, "bin", "ppm_ctl"), '-q', '-b', '-E', '-f', os.path.join(rootpath, "etc", "pleskd.props"), 'add', 'pkgmass']
	pkglist = __filterCustomPackages(pkglist)

	if topdir is not None:
		command += [ os.path.join(topdir, x.manifest_file) for x in pkglist]
	else:
		command += [ os.path.join(x.topdir, x.manifest_file) for x in pkglist]

	debug("doMassImport: %s, %s", command, topdir)

	if sys.platform != 'win32':
		p = sp.Popen(command, bufsize = 1, stdout=sp.PIPE, stderr=sp.PIPE)
	else:
		p = sp.Popen(command, bufsize = 1, stdout=sp.PIPE, stderr=sp.PIPE, startupinfo = uUtil.startup_info)
		
	out_thread = threading.Thread(target = __assignPkgIds, args = (p.stdout, pkglist))
	out_thread.setDaemon(True)
	errlines = []
	err_thread = threading.Thread(target = __showProgress, args = (p.stderr, pkg_counter, errlines))
	err_thread.setDaemon(True)
	out_thread.start()
	err_thread.start()
	out_thread.join()
	err_thread.join()
	status = p.wait()

	if status > 0:
		raise Exception('ppm ctl exited with code %d: %s' % (status, "".join(errlines)))
	elif status < 0:
		raise Exception('ppm ctl terminated with signal %d' % -status)


def recover():
	con = uSysDB.connect()
	cur = con.cursor()

	cur.execute("SELECT pkg_id, platform_id FROM packages WHERE name='pleskd' AND ctype='other' AND version='0'")

	bad_pleskds = [ (row[0], row[1]) for row in cur.fetchall() ]

	for pkg_id, platform_id in bad_pleskds:
		warn("Previous update attempt was interrupted, fixing pleskd package %d for platform %d", pkg_id, platform_id)
		cur.execute("DELETE FROM packages WHERE name='pleskd' AND ctype='other' AND platform_id = %s AND pkg_id != %s", platform_id, pkg_id)
		cur.execute("UPDATE packages SET version='1.0' WHERE pkg_id = %s", pkg_id)
	con.commit()
	

def updateManifestSources(plesk_root, pkglist, pkg_counter, topdir = None):
	progress.do("updating agent manifests")
	con = uSysDB.connect()
# Update old package set its version to zero (zero is less than anything else :)
# After it, import new package, update components and properties to point to new version, and delete old one.
	for pkg in pkglist:
		progress.do("updating %s", pkg)
		old_pkg_id = pkg.pkg_id
		cur = con.cursor()
		cur.execute("UPDATE packages SET version = '0', filename = NULL WHERE pkg_id = %s", old_pkg_id)
		con.commit()
		doMassImport(plesk_root, [pkg], pkg_counter, topdir)

		cur.execute("UPDATE components SET pkg_id = %s WHERE pkg_id = %s", pkg.pkg_id, old_pkg_id)

		cur.execute("""
		   SELECT p.prop_id, p.name, p2.prop_id AS new_prop_id,
		   cp.component_id FROM component_properties cp
		   JOIN properties p ON (cp.prop_id = p.prop_id)
		   LEFT JOIN properties p2 ON (p.name = p2.name)
		   WHERE p.pkg_id = %s
		   AND p2.pkg_id = %s """,(old_pkg_id, pkg.pkg_id))

		rows = cur.fetchall()

		for row in rows:
			old_prop_id, prop_name, new_prop_id, component_id = row[0], row[1], row[2], row[3]
			if new_prop_id is None:
				debug("Deleting property %s (old prop_id %s)", prop_name, old_prop_id)
				cur.execute("""DELETE FROM component_properties
		                    WHERE component_id = %s AND prop_id = %s """, (component_id, old_prop_id))
			else:
				debug('Copying property %s (old prop_id %s, new prop_id %s)', prop_name, old_prop_id, new_prop_id)
				cur.execute("""UPDATE component_properties
				    SET prop_id = %s
		                    WHERE component_id = %s
		                    AND prop_id = %s """, (new_prop_id, component_id, old_prop_id))
		cur.execute("DELETE FROM packages WHERE pkg_id = %s", old_pkg_id)
		con.commit()
		progress.done()
		pkg_counter.new_item(pkg)
	progress.done()


def importPackages(newestPackages, plesk_root, topdir = None, pkg_counter = None):
	# import pleskds - it registers platforms in system.
	progress.do("importing new packages manifests")
	new_pleskds = []

	total_packages = 0
	for pform_ctypes in newestPackages.values():
		for ctype_packages in pform_ctypes.values():
			total_packages += len(ctype_packages)

	if not pkg_counter:
		pkg_counter = uUtil.CounterCallback()

	pkg_counter.set_total(total_packages)

	for pform in newestPackages:
		if newestPackages[pform].has_key('other') and newestPackages[pform]['other'].has_key('pleskd'):
			new_pleskds.append(newestPackages[pform]['other']['pleskd'])
	
	if new_pleskds:
		progress.do("importing new agent packages")
		retriable(doMassImport)(plesk_root, new_pleskds, pkg_counter, topdir)
		to_update = [ x for x in new_pleskds if x.old ]
		if to_update:
			updateManifestSources(plesk_root, to_update, pkg_counter, topdir)
		progress.done()

	for pform in newestPackages:
		progress.do("importing new packages for %s", pform)
		for ctype in newestPackages[pform]:
			if newestPackages[pform][ctype]:
				progress.do("importing %s packages for %s", ctype, pform)
				retriable(doMassImport)(plesk_root, newestPackages[pform][ctype].values(), pkg_counter, topdir)
				progress.done()
		progress.done()
	progress.done()

def _transferFiles(files, host_id, path):
	MN_host_id = 1
	r = uHCL.Request(host_id)
	r.transfer(MN_host_id, [(f, path) for f in files])
	r.perform()

def updateOneMirror(host, path, tarballs_to_copy, host_id):
	progress.do("updating packages at mirror '%s'", host)
	if host_id == 1:
		for fn in tarballs_to_copy:
			uPEM.ln_cp(fn, path)
			print fn,
			sys.stdout.flush()
	else:
		for index in xrange(0, len(tarballs_to_copy), 100):
			portion = tarballs_to_copy[index:(index + 100)]
			_transferFiles(portion, host_id, path)

	progress.done()

def getLatestPackageId(pform_id, pkg):
	con = uSysDB.connect()
	cur = con.cursor()
	ctype, name = pkg
	cur.execute("SELECT pkg_id, version FROM packages WHERE platform_id = %s AND ctype = %s AND name = %s", pform_id, ctype, name)
	rv = None
	for row in cur.fetchall():
		pkg_id, version = row[0], row[1]
		if rv is None or compare_versions(version, rv[1]) > 0:
			rv = pkg_id, version

	if rv is not None or pform_id == 1:
		return rv
	cur.execute("SELECT parent_id FROM platforms WHERE platform_id = %s", pform_id)
	return getLatestPackageId(cur.fetchone()[0], pkg)

class NativeRepository:
	def __init__(self, id, url, platform, host_id, local_path):
		self.repo_id = id
		self.url = url
		self.host_id = host_id
		self.platform = platform
		self.local_path = local_path

# Retrive non-local managed and local managed located on MN native repositories
def getNativeRepos():
	con = uSysDB.connect()
	cur = con.cursor()
	cur.execute( \
		#	0		1		2	3	4	5	6
		"SELECT nr.repo_id, nr.local_path, p.opsys, p.osrel, p.arch, u.url, nr.host_id "\
		"FROM native_repositories nr " \
		"JOIN platforms p ON (nr.platform_id = p.platform_id) " \
		"JOIN ppm_urls u ON (u.url_id = nr.url_id) WHERE " \
		"pkg_manager_id = 0 AND " \
		"(ppm_mirror IS NOT NULL) OR (ppm_mirror IS NULL AND host_id = 1) " \
		"ORDER BY nr.repo_id")

	return [
		NativeRepository(
			id = int(x[0]),
			local_path = x[1],
			platform = Platform(x[2], x[3], x[4]),
			url = x[5],
			host_id = x[6])
		for x in cur.fetchall() ]


rpm_pattern = re.compile(r"(.+)\.([^.]+).rpm")
x86_to_both = [ r'postgresql90-libs-.*', r'h2e-webserver.*', r'pleskd-.*' ]
x86_to_both_patterns = [ re.compile(q) for q in x86_to_both ]

def fillOneRepo(build, repo, iprogress, pstart, ptotal):
	located_on_MN = repo.host_id == 1
	pform = repo.platform
	main_opsys = pform.os
	main_osver = pform.osver

	if pform.os == 'CentOS':
		if pform.osver == '4':
			main_opsys = 'RHES'
		else:
			main_opsys = 'RHEL'
	if pform.os == 'PSBM':
		main_opsys = 'RHEL'
		main_osver = '6'
	rpm_files = build.rpm_files
	topdir = build.topdir
	
	correct_rpm_files = {}
	x86_64 = 'x86_64'
	to_remove = []

	for rpm in rpm_files:
		m = rpm_pattern.match(rpm)
		if not m:
			err("%s: bad RPM file name", rpm)
			correct_rpm_files[rpm] = rpm_files[rpm]
			continue
		rpm_name = m.group(1)
		rpm_arch = m.group(2)
		if pform.arch == x86_64:
			# we need to copy some i386 rpms to x86_64 repo at least for pleskd, which is still x86
			do_copy = rpm_arch == x86_64 or first(x86_to_both_patterns, lambda p: p.match(rpm)) or (not rpm_files.has_key('%s.x86_64.rpm' % rpm_name))
		else:
			do_copy = rpm_arch != x86_64

		if do_copy:
			correct_rpm_files[rpm] = rpm_files[rpm]
		else:
			to_remove.append(rpm)

	os_rpms = lambda o: [ os.path.join(topdir, rpm_files[x][o], x) for x in correct_rpm_files.keys() if o in rpm_files[x] ]
	pkgs_path = os_rpms((pform.os, pform.osver))
	if pform.os == 'CentOS' or pform.os == 'PSBM':
		pkgs_path += os_rpms((main_opsys, main_osver))

	pkgs_path += os_rpms(None)

	# effective for remote repos only
	to_transfer = []

	for i, p in enumerate(pkgs_path):
		if located_on_MN:
			uPEM.ln_cp(p, repo.local_path)
		else:
			to_transfer.append([os.path.abspath(p), repo.local_path])

		if iprogress:
			iprogress.set_progress(pstart + int(ptotal * i / len(pkgs_path)))

	if not located_on_MN:
		debug("Transfer rpm files to remote native repository")
		hcl = uHCL.Request(repo.host_id)
		hcl.transfer(1, to_transfer)
		hcl.perform()
	
	if located_on_MN:
		# the following tries to fix incorrectly filled repository (arch-mess)
		for fn in to_remove:
			fn = os.path.join(repo.local_path, fn)
			if os.path.isfile(fn):
				debug("Removing %s, should not be here", fn)
				try:
					os.unlink(fn)
				except:
					pass

def fillNativeRepositories(builds, pprogress, iprogress = None, pstart = 0, ptotal = 100):
	pprogress.do("updating native repositories")
	nr = getNativeRepos()
	to_reindex = []

	per_build = (ptotal - 30) / len(builds)
	for bn, build in enumerate(builds):
		for rep_no, r in enumerate(nr):
			pprogress.do("updating repository %s at %s", r.repo_id, r.local_path)
			if iprogress:
				iprogress.set_progress(pstart + bn * per_build + per_build * rep_no / len(nr), "Updating repository %s at %s" % (r.repo_id, r.local_path))
			start = pstart + bn * per_build + per_build * rep_no / len(nr)
			fillOneRepo(build, r, iprogress, start, 1.0 * per_build / len(nr))

			to_reindex.insert(len(to_reindex), r.repo_id)
			pprogress.done()

	pprogress.do("reindexing native repositories")
	nr_mgr = uNativeRepo.NativeRepoManager()
	for i, repo_id in enumerate(to_reindex):
		if iprogress:
			iprogress.set_progress(pstart + ptotal - 30 + 30 * i / len(to_reindex), "Reindexing repository %s" % (repo_id, ))
		nr_mgr.index_repo({ 'repo_id' : repo_id })
	pprogress.done()
	pprogress.done()
		

def schedulePackagesUpgrade(packages, api, depends = None):
	if depends is None:
		depends = []
	to_install = []
	for pkg in packages:
		hosts = getPackageHosts(pkg)
		for pform_id in hosts:
			pkg_id = getLatestPackageId(pform_id, pkg)
			pform = uPEM.getPlatform(pform_id, uSysDB.connect())
			if pkg_id is None:
				err("%s: cannot find package for platform %s", pkg, pform)
				continue
			pkg_id, version = pkg_id
			debug("Going to install %s v %s (%s) at %s", pkg, version, pkg_id, pform)
			to_install += [ { 'host_id' : host_id, 'pkg_id': pkg_id } for host_id in hosts[pform_id]]
	
	return api.pem.packaging.asyncUpgrade(component_list = to_install, depend_on = depends)["task_ids"]

def scheduleOrderedUpgrade(packages):
	task_ids = []
	api = openapi.OpenAPI()
	api.begin()
	for line in packages:
		task_ids = schedulePackagesUpgrade(line, api, task_ids)
	api.commit()

def scheduleUpgradeSeries(packages):
	api = openapi.OpenAPI()
	pkgs = [ { 'ctype' : x[0], 'name' : x[1] } for x in packages ]
	api.pem.packaging.scheduleUpgradeSeries(packages = pkgs)

def syncMirrors(ppm_mirrors, newestPackages):
	progress.do("updating packages on PPM mirrors")

	tarballs_to_copy = []
	for pform in newestPackages:
		for ctype in newestPackages[pform]:
			tarballs_to_copy += [ x.tarball_location for x in newestPackages[pform][ctype].values() if x.tarball_location ]

	for m in ppm_mirrors:
		updateOneMirror(m.hostname, m.localpath, tarballs_to_copy, m.host_id)
	progress.done()

def getPackageHosts(pkg):
	ctype, name = pkg
	con = uSysDB.connect()
	cur = con.cursor()
	cur.execute("SELECT h.platform_id, h.host_id FROM hosts h JOIN components c ON (c.host_id = h.host_id) JOIN packages p ON (p.pkg_id = c.pkg_id) WHERE p.name = %s AND p.ctype = %s", name, ctype)
	rv = {}
	for row in cur.fetchall():
		pform_id = row[0]
		host_id = row[1]
		if rv.has_key(pform_id):
			rv[pform_id].append(host_id)
		else:
			rv[pform_id] = [ host_id ]

	return rv


def getMainMirror():
	con = uSysDB.connect()
	cur = con.cursor()

	cur.execute("SELECT m.local_path, u.url FROM ppm_mirrors m "
		"JOIN ppm_urls u ON (m.url_id = u.url_id) WHERE "
		"m.host_id = 1 ORDER BY CASE WHEN m.service_id IS NULL THEN -1 ELSE m.service_id END")

	row = cur.fetchone();

	if not row:
		return None
	return row[0], row[1]

class PPMMirror:
	def __init__(self, id, hostname, localpath, host_id, url):
		self.id = id
		self.hostname = hostname
		self.localpath = localpath
		self.host_id = host_id
		self.url = url

def getPPMMirrors():
	con = uSysDB.connect()
	cur = con.cursor()
	cur.execute(
		"SELECT h.primary_name, pm.local_path, pm.host_id, u.url, pm.mirror_id " 
		"FROM ppm_mirrors pm "
		"JOIN hosts h ON (pm.host_id = h.host_id) "
		"JOIN ppm_urls u ON (u.url_id = pm.url_id)")
	
	ppm_mirrors = [
		PPMMirror(id = x[4], hostname = x[0], localpath = x[1], host_id = x[2], url = x[3]) for x in cur.fetchall()
	]

	if not ppm_mirrors and sys.platform != 'win32':
		raise Exception("No PPM mirrors configured. Please configure one.")
	elif sys.platform == 'win32':
		cur.execute("SELECT u.url FROM ppm_urls u JOIN host_urls hu ON (u.url_id = hu.url_id) WHERE hu.host_id = 1")
		row = cur.fetchone()
		if row:
			url = row[0]
			cur.execute("SELECT url FROM ppm_urls WHERE url LIKE 'http%' ORDER BY url_id")
			row2 = cur.fetchone()

			if row2:
				real_url = row2[0]
			else:
				real_url = url

			if url.startswith("file:///"):
				localpath = localpath=url[8:]
			elif url.startswith("file://"):
				localpath = localpath=url[7:] 

			ppm_mirrors.append(PPMMirror(id=None, hostname='Management node', localpath=localpath, host_id=1, url=real_url))
		if not ppm_mirrors:
			raise Exception("Failed to find any suitable ppm url")

	return ppm_mirrors

def clean_interfaces():
	con = uSysDB.connect()
	cur = con.cursor()
	cur.execute("DELETE FROM service_references WHERE NOT EXISTS (SELECT 1 FROM package_interfaces WHERE package_interfaces.interface_id = service_references.interface_id)")
	cur.execute("DELETE FROM interfaces WHERE NOT EXISTS (SELECT 1 FROM package_interfaces WHERE package_interfaces.interface_id = interfaces.interface_id UNION ALL SELECT 1 FROM dep_interfaces di WHERE di.interface_id = interfaces.interface_id)")
	con.commit()
	cur.execute("SELECT p.pkg_id, p.name, p.ctype, i.service_type FROM interfaces i JOIN dep_interfaces di ON (i.interface_id = di.interface_id) JOIN package_dependencies dp ON (di.dep_id = dp.dep_id) JOIN packages p ON (dp.pkg_id = p.pkg_id) WHERE i.interface_id NOT IN (SELECT interface_id FROM package_interfaces)")
	pkgs = {}
	for row in cur.fetchall():
		pkg_name = row[1], row[2]
		if not pkgs.has_key(pkg_name):
			pkgs[pkg_name] = [], []
		pkg_id = str(row[0])
		iface = row[3]
		if pkg_id not in pkgs[pkg_name][0]:
			pkgs[pkg_name][0].append(pkg_id)
		if iface not in pkgs[pkg_name][1]:
			pkgs[pkg_name][1].append(iface)

	for pkg in pkgs:
		warn("%s-%s (%s) depends on non-existing interface(s): %s",
			pkg[0], pkg[1], ', '.join(pkgs[pkg][0]), ', '.join(pkgs[pkg][1]))

__all__ = [ 'installPackageToHost', 'installPackageToHostAPI', 'removePackage', 'compare_versions', 'registerPlatform', "getNewestPackages", "doMassImport", "importPackages", "syncMirrors", "getPPMMirrors", "getNativeRepos" ]
