import sys
import exceptions
import time
import uLogging

paramstyle='pyformat'
apilevel="2.0"
threadsafety="1"


class Warning(exceptions.StandardError):
	def __init__(self, *args):
		exceptions.StandardError.__init__(self, *args)

class Error(exceptions.StandardError):
	def __init__(self, *args):
		exceptions.StandardError.__init__(self, *args)

class DatabaseError(Error):
	def __init__(self, *args):
		Error.__init__(self, *args)


class OperationalError(DatabaseError):
	def __init__(self, *args):
		DatabaseError.__init__(self, *args)

	
class NotSupportedError(DatabaseError):
	def __init__(self, *args):
		DatabaseError.__init__(self, *args)


class Cursor:
	def _update_attrs(self):
		if hasattr(self.cursor, 'description'):
			self.description = self.cursor.description
		else:
			self.description = None

		if hasattr(self.cursor, 'rowcount'):
			self.rowcount = self.cursor.rowcount
		else:
			self.rowcount = None

	def close(self):
		self.cursor.close()
		del self.cursor
	
	def callproc(self, *args):
		raise NotSupportedError("No stored procedures (%s), please", args)

	def __init__(self, cursor):
		self.cursor = cursor
	
	def fetchone(self):
		rv = self.cursor.fetchone()
		return rv

	def fetchmany(self, *args):
		rv = self.cursor.fetchmany(*args)
		return rv

	def fetchall(self):
		rv = self.cursor.fetchall()
		return rv

	def nextset(self):
		raise NotSupportedError("Multiple resultsets are not supported")

	def setinputsizes(self, *args):
		pass
	
	def setoutputsize(self, *args):
		pass
	

class DictConverter:
	# a bit hackerish. It assumes that % formatting calls []
	# exactly once for each converting. It does so.
	def __init__(self):
		self.argmap = {}
		self.arglist = []
	
	def convert(self, format, argmap):
		self.arglist = []
		self.argmap = argmap

		return format % self.argmap, tuple(self.arglist)
	
	def __getitem__(self, key):
		self.arglist.append(self.argmap[key])
		return '?'


def convert_arg(arg):
	return arg

class QmarkCursor(Cursor):
	def __init__(self, cursor):
		Cursor.__init__(self, cursor)
	
	def execute_org(self, statement, *args):
		if len(args) == 0:
			rv = self.cursor.execute(statement)
		elif len(args) != 1:
			# parameters is sequence
			c_args = [ convert_arg(x) for x in args]
			rv = self.cursor.execute(statement % (('?',) * len(c_args)), c_args)
		elif type(args[0]) == list or type(args[0]) == tuple :
			# parameters are sequence ==> positional substitution
			# just replace %s with ?.
			c_args = [ convert_arg(x) for x in args[0]]
			rv = self.cursor.execute(statement % (('?',) * len(c_args)), c_args)
		elif type(args[0]) == dict:
			rv = self.cursor.execute(*(DictConverter().convert(statement, args[0])))
		else: 
			# one parameter
			rv = self.cursor.execute(statement % '?', (convert_arg(args[0]),))

		return rv
	def execute(self, statement, *args):
		try :
			return self.execute_org(statement, *args)
		except :
			exctype, excvalue, traceback = sys.exc_info()
			if type(exctype) == str:
				raise DatabaseError(exctype, excvalue, "while executing %s (%s)" % (statement, args))
			raise

	def fetchone(self):
		try:
			rv = self.cursor.fetchone()
			return rv
		except :
			exctype, excvalue, traceback = sys.exc_info()
			if type(exctype) == str:
				raise DatabaseError(exctype, excvalue, "while fetching one row")
			raise

	
class PyFormatCursor(Cursor):
	def __init__(self, cursor):
		Cursor.__init__(self, cursor)
	
	def execute(self, statement, *args):
		try:
			if not args:
				rv = self.cursor.execute(statement)
			elif len(args) == 1 and type(args[0]) in (tuple, list):
				rv = self.cursor.execute(statement, args[0])
			else:
				rv = self.cursor.execute(statement, args)
		except NativeOperationalError, e:
			e_args = [ "executing stmt with params:", statement, args] + list(e.args)
			raise OperationalError(e_args)
			
		except Exception, e:
			print e, e.args
			e_args = [ "executing stmt with params:", statement, args] + list(e.args)
			raise DatabaseError(e_args)
		return rv


all_connections = {}

class Connection:
	def __init__(self, connection):
		self.connection = connection

	def close(self):
		self.connection.rollback()

	def commit(self):
		return self.connection.commit()

	def rollback(self):
		return self.connection.rollback()

	def insertRecord(self, table, **kwds):
		colstr = ""
		params = []
		for k in kwds:
			if colstr:
				colstr += ', '
			colstr += k
			params.append(kwds[k])
		cur = self.cursor()
		stmt = "INSERT INTO %s (%s) VALUES (%s)" % (table, colstr, ', '.join(['%s'] * len(params)))

		cur.execute(stmt, params)

	def insertRecordWithId(self, table, **kwds):
		self.insertRecord(table, **kwds)
		return get_last_inserted_value(self, table)

class PyFormatConnection(Connection):
	def __init__(self, connection):
		Connection.__init__(self, connection)

	def cursor(self, *args, **kwds):
		return PyFormatCursor(self.connection.cursor(*args, **kwds))

	def commit(self):
		try:
			self.connection.commit()
		except _dbmodule.libpq.Warning, e:
			uLogging.warn("%s", e)
			
			

class QmarkConnection(Connection):
	def __init__(self, connection):
		Connection.__init__(self, connection)

	def cursor(self, *args, **kwds):
		return QmarkCursor(self.connection.cursor(*args, **kwds))

DBType = None

PgSQL = 1
MSSQL = 2

ConcatOperator = None
nowfun = None

NativeOperationalError = KeyboardInterrupt

_hamn_schema = None

def hamn_schema():
	global _hamn_schema
	return _hamn_schema

def set_hamn_schema(val):
	global _hamn_schema
	_hamn_schema = val

# XXX: not very correct - Win platform ==> MSSQL, it may be changed
if sys.platform == 'win32':
	import pyodbc

	def connect_win(sc = "plesk", super_user=False):
		if sc == "plesk":
			database = "plesk"
			uname = _config.database.user
			pwd = _config.database.password
		else:
			database = _databases[sc]
			uname, pwd = _database_users[sc]

		if super_user:
			uname = 'sa'

		if not all_connections.has_key((database, uname, pwd)):
			connect_string = "DSN=%s;UID=%s;PWD=%s" % (database, uname, pwd)
			all_connections[(database, uname, pwd)] = QmarkConnection(pyodbc.connect(connect_string))
		return all_connections[(database, uname, pwd)]

	connect = connect_win
	DBType = MSSQL
	ConcatOperator = '+'

	def toRaw(x):
		if x is None:
			return x
		else:
			return pyodbc.Binary(x)

	def get_last_inserted_value_mssql(con, tablename):
		cur = con.cursor()
		try :
			cur.execute("SELECT @@IDENTITY")
			return cur.fetchone()[0]
		finally:
			cur.close()
	get_last_inserted_value = get_last_inserted_value_mssql
	nowfun = 'getdate'

	convertDTstring = lambda t: time.strftime('%Y-%m-%d %H:%M:%S', time.strptime(t))

else:
	from pyPgSQL import PgSQL as _dbmodule
	NativeOperationalError = _dbmodule.libpq.OperationalError
	def toRaw(x):
		if x is None:
			return x
		else:
			return _dbmodule.PgBytea(x)
	convertDTstring = lambda x: x

	def connect_unix(sc = "plesk", super_user=False):
		if sc == "plesk":
			database = "plesk"
		else:
			database = _databases[sc]
		if super_user:
			uname = 'postgres'
		else:
			uname = database
		if not all_connections.has_key((database, uname)):
			all_connections[(database, uname)] = PyFormatConnection(_dbmodule.connect(host=_config.database.host, user=uname, database=database))

		rv = all_connections[(database, uname)]
		try:
			cur = rv.cursor()
			cur.execute("SELECT 1")
		except OperationalError:
			uLogging.warn("Connection to DB is lost, attempting to reconnect");
			rv = all_connections[(database, uname)] = PyFormatConnection(_dbmodule.connect(host=_config.database.host, user=uname, database=database))

		return rv
	
	def get_last_inserted_value_pgsql(con, tablename):
		cur = con.cursor()
		cur.execute("SELECT CURRVAL('%s_seq')" % tablename)
		return cur.fetchone()[0]
		
	connect = connect_unix

	get_last_inserted_value = get_last_inserted_value_pgsql

	DBType = PgSQL
	ConcatOperator = '||'
	nowfun = 'now'

_databases = {}
_database_users = {}
_config = None

def table_exist(con, table_name):
	global DBType
	if DBType == PgSQL:
		query = "SELECT 1 FROM pg_class WHERE relname = %s"
	else:
		query = "SELECT 1 FROM sys.objects WHERE name = %s"
	cur = con.cursor()

	cur.execute(query, table_name)
	if cur.fetchone():
		return True
	else:
		return False

def sc_database(sc):
	if sc == 'plesk':
		return 'plesk'
	elif sc in _databases:
		return _databases[sc]
	else:
		return None


def rollback_all():
	for i in all_connections:
		all_connections[i].rollback()

def commit_all():
	for i in all_connections:
		all_connections[i].commit()

def init(config=None, connection = None):
	global _config
	global _databases
	global _database_users
	import os

	if config is not None:
		_config = config
	if _config is None:
		raise Exception, "config arg is mandatory when you're initializing uSysDB first time"

	_databases = {}
	if connection is None:
		con = connect()
	else:
		con = connection
	os.environ['hostname'] = _config.database.host
	
	os.environ['plesk_db'] = 'plesk'
	if not table_exist(con, "service_classes"):
		return

	set_hamn_schema(table_exist(con, 'sc_instances'))

	if hamn_schema():
		q = "SELECT DISTINCT sc.name, p.value FROM v_props p JOIN sc_instances si ON (p.component_id = si.component_id) JOIN service_classes sc ON (sc.sc_id = si.sc_id) WHERE p.name='dsn' AND p.value != ''" 
	else:
		q = "SELECT sc.name, p.value FROM v_props p JOIN service_classes sc ON (p.component_id = sc.component_id) WHERE p.name='dsn' AND p.value != ''"

	cur = con.cursor()
	cur.execute(q)
	for row in cur.fetchall():
		_databases[row[0]] = row[1]
		os.environ['%s_db' % row[0]] = row[1]

	if hamn_schema():
		q = "SELECT sc.name, pl.value, pp.value FROM service_classes sc JOIN sc_instances si ON (sc.sc_id = si.sc_id) JOIN v_props pl ON (pl.component_id = si.component_id AND pl.name = 'dsn_login') JOIN v_props pp ON (pp.component_id = si.component_id AND pp.name = 'dsn_passwd')"
	else:
		q = "SELECT sc.name, pl.value, pp.value FROM service_classes sc JOIN v_props pl ON (pl.component_id = sc.component_id AND pl.name = 'dsn_login') JOIN v_props pp ON (pp.component_id = sc.component_id AND pp.name = 'dsn_passwd')"

	cur.close()
	cur = con.cursor()
	cur.execute(q)
	for row in cur.fetchall():
		_database_users[row[0]] = row[1], row[2]

	cur.execute("SELECT name FROM service_classes WHERE name NOT IN ('pleskd', 'vzpemagent')")
	
	for row in cur.fetchall():
		scname = row[0]
		if scname not in _databases:
			_databases[scname] = _databases['Kernel']
		if scname not in _database_users:
			_database_users[scname] = _database_users['Kernel']


	cur.close()
	if connection is None:
		con.close()



__all__ = [ "connect", "table_exist", "sc_database", "get_last_inserted_value", 'paramstyle', 'apilevel', 'threadsafety', 'DBType', 'MSSQL', 'PgSQL', 'toRaw', 'ConcatOperator', 'hamn_schema', 'set_hamn_schema' ]
