__rcs_id__ = """$Id: uTasks.py 579920 2010-06-03 13:46:45Z csvintsoff $"""
__pychecker__ = "blacklist=modules/sets.py unusednames=__rcs_id__,dummy"

import uSysDB
from uLogging import *
import uPEM
import openapi
import uActionContext

class SCRef:
	def __init__(self, service_type, version = "", sc_id = 0):
		self.service_type = service_type
		self.version = version
		self.sc_id = int(sc_id)

	def to_api(self):
		return {
			"service_type" : self.service_type,
			"sc_id" : self.sc_id
		}

	def to_db(self):
		if self.version:
			return ('SCREF', "%s:%s:%d" % (self.service_type, self.version, self.sc_id))
		else:
			return ('SCREF', "%s:%d" % (self.service_type, self.sc_id))


class PObjRef:
	def __init__(self, scref, fun, obj_id):
		self.scref = scref
		self.fun = fun
		self.obj_id = int(obj_id)

	def to_api(self):
		rv = self.scref.to_api()
		rv["get_method"] = self.fun
		rv["obj_id"] = self.obj_id
		return rv

	def to_db(self):
		return ('OBJREF', "%s:%s:%d" % (self.scref.to_db()[1], self.fun, self.obj_id))


def daily(hours, minutes, period = 1):
	return ('d', hours, minutes, None, period)

def weekly(day, hours, minutes, period = 1):
	return ('w', hours, minutes, day, period)

def monthly(day, hours, minutes, period = 1):
	return ('m', hours, minutes, day, period)

def periodically(period):
	return ('s', None, None, None, period)


class DBTaskManagement:
	def __init__(self, con):
		self.con = con

	def scheduleUsual(self, task):
		debug("DBTaskManagement.scheduleUsual %s" % task.name)
		task_id = self.__scheduleCommon(task)
		cur = self.con.cursor()
		cur.execute("INSERT INTO tm_usual(task_id, mutex, ignore_failures, retry_num) VALUES(%s, %s, 'n', %s)", (task_id, task.mutex, task.retries))
		for p in task.params:
			cur.execute("INSERT INTO tm_params(task_id, p_name, p_value) VALUES (%s, %s, %s)", task_id, p, task.params[p])
		return task_id

	def schedulePeriodic(self, task):
		debug("DBTaskManagement.schedulePeriodic %s" % task.name)
		task_id = self.__scheduleCommon(task)
		cur = self.con.cursor()
		cur.execute("INSERT INTO tm_periodic(task_id, period_type, start_at_hour, start_at_min, start_at_day, period) VALUES(%s, %s, %s, %s, %s, %s)", (task_id,) + task.period)
		return task_id

	def __scheduleCommon(self, task):
		cur = self.con.cursor()
		cur.execute(("INSERT INTO tm_tasks(name, description, location, method, next_start, status) VALUES(%s, %s, %s, %s, "+ uSysDB.nowfun + "(), 'u')"),
			(task.name, task.description, "%s:%s" % task.ref.to_db(), task.method))
		return uSysDB.get_last_inserted_value(self.con, "tm_tasks")

class APITaskManagement:
	def __init__(self):
		action_id = uActionContext.action_id()

		if action_id is not None:
			txn_id = "uTasks.%s" % action_id
			self.api = uActionContext.get(txn_id)

			if self.api is None:
				self.api = openapi.OpenAPI()
				self.api.begin(txn_id)
				uActionContext.put(txn_id, self.api)
		else:
			# Do not use Open API transactions outside of actions
			self.api = openapi.OpenAPI()

	def scheduleUsual(self, task):
		debug("APITaskManagement.scheduleUsual %s" % task.name)
		args = {}

		if task.mutex:
			args['mutex'] = task.mutex

		if task.retries > 0:
			args['retries'] = { 'number' : task.retries }

		if task.params:
			args['params'] = [ { 'name' : str(k), 'value' : str(v) } for k, v in task.params.items() ]

		return self.__scheduleCommon(task, args)

	def schedulePeriodic(self, task):
		debug("APITaskManagement.schedulePeriodic %s" % task.name)
		fn = {
			's' : lambda x: { 'period_seconds' : int(x[3]) },
			'd' : lambda x: { 'daily'   : { 'hour' : int(x[0]), 'minute' : int(x[1]) } },
			'w' : lambda x: { 'weekly'  : { 'hour' : int(x[0]), 'minute' : int(x[1]), 'day' : int(x[2]) } },
			'm' : lambda x: { 'monthly' : { 'hour' : int(x[0]), 'minute' : int(x[1]), 'day' : int(x[2]) } },
		}

		p = task.period
		parg = fn[p[0]](p[1:])

		if p[0] != 's':
			parg['period'] = p[4]

		args = { 'periodic' :  parg }

		return self.__scheduleCommon(task, args)

	def __scheduleCommon(self, task, more_args):
		args = {
			'processor'	: task.ref.to_api(),
			'name'		: task.name,
			'description'	: task.description
		}
		args['processor']['method'] = task.method
		args.update(more_args)
		return int(self.api.pem.tasks.schedule(**args)['task_id'])

class TaskBase:
	def __init__(self, ref, name, method, description = None):
		self.ref = ref
		self.name = name
		self.method = method

		if description is None:
			self.description = name
		else:
			self.description = description

	def get_sched_strategy(self, con):
		if con is None:
			con = uSysDB.connect()

		if uPEM.is_started('TaskManager') and uPEM.is_started('OpenAPI'):
			return APITaskManagement()
		else:
			return DBTaskManagement(con)

class PeriodicTask(TaskBase):
	def __init__(self, ref, name, method, period, description = None):
		TaskBase.__init__(self, ref, name, method, description)
		self.period = period

	def schedule(self, con = None):
		return self.get_sched_strategy(con).schedulePeriodic(self)

class Task(TaskBase):
	def __init__(self, ref, name, method, mutex = None, description = None):
		TaskBase.__init__(self, ref, name, method, description)
		self.mutex = mutex
		self.params = {}
		self.retries = 0

	def setParam(self, param, value):
		self.params[param] = value

	def setRetryNum(self, num):
		self.retries = num
	
	def schedule(self, con = None):
		return self.get_sched_strategy(con).scheduleUsual(self)
